/*-----------------------------------------
   Calc.C -- Calculation module
   Version date: 03.09.22
  -----------------------------------------*/

/*----------------------------------------------------------------------
COPYRIGHT 2003, 2005 PETER DAVIS
This file is part of "HMS Surprise - Frigate and brig Simulator for
Windows". That program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version. The file is distributed
in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this file; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
----------------------------------------------------------------------*/

#include <windows.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "External.h"

#define SAILSTEP           0.5

// Prototypes:
void          Calculate () ;
void          ReplaySailStatus (int) ;

extern void   CanShadow () ;
extern void   SailPlan () ;
extern void   SetUpDisplay () ;
extern void   ShowMessage () ;
extern int    Sgn (float) ;

/* --------------------------------------
   Calculate: performs all cyclical calculations
   Version date = 03.09.22
----------------------------------------*/

void Calculate ()
  {
  float        fAlongShip ;
  float        fAngle ;
  float        fArea ;
  char         sBuf [20];
  float        fCrossShip ;
  float        fDeltaHdg ;
  float        fDeltaSpeed ;
  float        fDeltaLat, fDeltaLon ;
  float        fDist ;
  float        fDistance ;
  float        fEffPerpArea ;     
  float        fFactor ;
  float        fH ;
  int          iHours ;
  int          iI, iJ, iK ;
  int          i1, i2 ;
  static int   iLastCOEHeight ;
  static int   iLastCOELeeward ;
  static int   iLastHeelingAngle ;
  float        fLeftEdge ;
  int          iMinutes ;
  float        fMoment ;
  int          iN ;
  float        fNewLeftEdge ;
  float        fOfset ;
  float        fParalax ;
  float        fRightEdge ;
  float        fSail ;
  int          iSeconds ;
  BOOL         bShad ;
  float        fShad ;
  float        fShadLeftEdge ;
  float        fShadRightEdge ;
  int          cSh ;
  int          iSign ;
  int          cSl ;
  float        fSlantLeft, fSlantRight ;
  float        fSlHeight ;
  float        fSpeedEast ;
  float        fSpeedNorth ;
  BOOL         bSplit ;
  float        fTemp, fTemp2 ;
  float        fTempL, fTempR ;
  float        fTotalPerpArea ;
  float        fTotalThrust ;

  QueryPerformanceFrequency (&liFrequency);
  
  // clock
  if (liFrequency.QuadPart == 0)
    {
    fNow = (float) clock()/CLOCKS_PER_SEC ;
    }
  else
    {
    QueryPerformanceCounter (&liCounter) ;
    fNow = (float) (liCounter.QuadPart / (float) liFrequency.QuadPart) ;
    }
  fIncrement = (fNow - fLastTime) ;
  fLastTime = fNow ;
  fRealTime = fNow-fRecordStartTime ;
  fSimTime = fLastSimTime + (fRealTime-fLastRealTime) * iAcceleration[iSpeedFactor] ;
  fLastRealTime = fRealTime ;
  fLastSimTime = fSimTime ;

  // Calculate relative wind angle
  // negative = port, positive = starboard, 0 = astern, 180 = ahead
  fRelWindAngle = fHeading - (fWindDirection + 180.0) ;
  if (fRelWindAngle < -180.0)
    fRelWindAngle = fRelWindAngle + 360.0 ;
  if (fRelWindAngle > 180.0)
    fRelWindAngle = fRelWindAngle - 360.0 ;

  // Calculate apparent wind vector
  fAlongShip = fWindSpeed * cos(fRelWindAngle * fDegToRad) - fSpeed ;
  fCrossShip = fWindSpeed * sin(fRelWindAngle * fDegToRad) + fLeewaySpeed ;
  if (fabs(fAlongShip) <= 0.1)
    // avoid division by zero
    fAppWindAngle = 90 ;
  else
    fAppWindAngle = atan(fabs(fCrossShip/fAlongShip)) * 180.0 / fPi ;
  if (fAlongShip < 0 )
    fAppWindAngle = 180 - fAppWindAngle ;
  fAppWindAngle = fAppWindAngle * Sgn(fRelWindAngle) ;
  if (fabs(fAppWindAngle - fOldAppWindAngle) > 1)
    {
    //update PlanView
    fOldAppWindAngle = fAppWindAngle ;
    bPlanViewIsChanged = TRUE ;
    }
  fAppWindSpeed = sqrt(fCrossShip * fCrossShip + fAlongShip * fAlongShip) ;
  fAppWindDirection = fHeading - (180.0 + fAppWindAngle) ;
  if (fAppWindDirection < 0)
    fAppWindDirection = fAppWindDirection + 360.0 ;

  // Check if staysails are still pulling
  for (cSl = 0; cSl < iNrSails ; cSl++) 
    {
    if ((iSailType[cSl]==1) || (iSailType[cSl]==3))
      {
      if (Sgn(fSailAngle[cSl]) == -Sgn(fAppWindAngle))
        fSailAngle[cSl] = -fSailAngle[cSl] ; // sail flips over
      }
    }

  // Calculate the apparent distances of the masts and the corners of the sails
  //  from the centre of rotation when viewed from the direction of the wind
  fDist = sin(fAppWindAngle * fDegToRad) ;
  for (cSl = 0 ; cSl < iNrSails ; cSl++)
    {
    switch (iSailType[cSl])
      {
      case 2:
        // foremast sails
        fParalax = fDist * fMastDistance[0] ;
        break ;
      case 4:
        // mainmast sails
        fParalax = fDist * fMastDistance[1] ;
        break ;
      case 6:
        // mizzenmast sails
        fParalax = fDist * fMastDistance[2] ;
        break ;
      case 1:
      case 3:
        // stay sails
        fParalax = fDist * fSailTackDistance[cSl-(iSpanker+1)] ;
      }

    // Calculate Sailplan positions of corners of effective sail
    fTempL = xSailPlanHeadLeft[cSl] ;
    fTempR = xSailPlanHeadRight[cSl] ;
    if ((iSailType[cSl] == 2) || (iSailType[cSl] == 4) || (iSailType[cSl] == 6))
      {
      if (fSailEffectiveHead[cSl] != fSailHead[cSl])
        {
        fTemp = (fSailEffectiveHead[cSl] - fSailFoot[cSl]) * 2.0 ;
        fSlantLeft = (xSailPlanFootLeft[cSl] - xSailPlanHeadLeft[cSl]) / 
          ((fSailHead[cSl] - fSailFoot[cSl]) * 2.0) ;
        fSlantRight = (xSailPlanFootRight[cSl] - xSailPlanHeadRight[cSl]) / 
          ((fSailHead[cSl] - fSailFoot[cSl]) * 2.0) ;
        fTempL = xSailPlanFootLeft[cSl] - fSlantLeft * fTemp ;
        fTempR = xSailPlanFootRight[cSl] - fSlantRight * fTemp ;
        }
      }

    fFactor = (sin((fAppWindAngle + 90.0 + fSailAngle[cSl]) * fDegToRad)) ;
    if ((iSailType[cSl] == 1) || (iSailType[cSl] == 3) || (cSl == iSpanker))
       fFactor = (sin((fAppWindAngle + fSailAngle[cSl]) * fDegToRad)) ;

    xSailWindHeadLeft[cSl] = fParalax + fFactor * fTempL ;
    xSailWindHeadRight[cSl] = fParalax + fFactor * fTempR ;
    xSailWindFootLeft[cSl] = fParalax + fFactor * xSailPlanFootLeft[cSl] ;
    xSailWindFootRight[cSl] = fParalax + fFactor * xSailPlanFootRight[cSl] ;
    if ((iSailType[cSl] == 1) || (iSailType[cSl] == 3))
      // staysails
      {
      fFactor = (sin((fAppWindAngle + fSailAngle[cSl]) * fDegToRad)) ;
      xSailWindFootLeft[cSl] = fParalax + fFactor * xSailPlanFootLeft[cSl] ;
      fFactor = sin(fAppWindAngle * fDegToRad) ;
      xSailWindHeadLeft[cSl] = fParalax + fFactor * xSailPlanHeadLeft[cSl] ;
      xSailWindHeadRight[cSl] = fParalax + fFactor * xSailPlanHeadRight[cSl] ;
      xSailWindFootRight[cSl] = fParalax + fFactor * xSailPlanFootRight[cSl] ;
      }
    }

  // Determine which sails are reached first by the wind
  if (bBrig)
    {
    if (sin(((90.0 - fSailAngle[9]) - fAppWindAngle) * fDegToRad) >= 0.0)    // mainyard drawing
      {
      if (sin(((90.0 - fSailAngle[4]) - fAppWindAngle) * fDegToRad) >= 0.0)  // foreyard drawing
        iSeq = 1 ;
      else
        iSeq = 0 ;
      }
    else    // mainsail aback
      { 
      if (sin(((90.0 - fSailAngle[4]) - fAppWindAngle) * fDegToRad) >= 0.0)
        iSeq = 3 ;
      else
        iSeq = 2 ;
      }
    }
  else
    {
    if (sin(((90.0 - fSailAngle[7]) - fAppWindAngle) * fDegToRad) >= 0.0)    // mainyard drawing
      {
      if (sin(((90.0 - fSailAngle[3]) - fAppWindAngle) * fDegToRad) >= 0.0)  // foreyard drawing
        iSeq = 1 ;
      else
        iSeq = 0 ;
      }
    else    // mainsail aback
      { 
      if (sin(((90.0 - fSailAngle[3]) - fAppWindAngle) * fDegToRad) >= 0.0)
        iSeq = 3 ;
      else
        iSeq = 2 ;
      }
    }
 
  // Calculate the equations which define the (upper) left and right edges of the sails seen from the direction of the wind
  for (iI = 0; iI <= iNrSails; iI++)
    {
    if (xSailWindHeadLeft[iI]==xSailWindFootLeft[iI]) 
      {
      fSailGradientLeft[iI] = 0 ;
      fSailInterceptLeft[iI] = 0 ;
      }
    else
      {
      fSailGradientLeft[iI] = (ySailPlanHeadLeft[iI]-ySailPlanFootLeft[iI])/(xSailWindHeadLeft[iI]-xSailWindFootLeft[iI]) ;
      fSailInterceptLeft[iI] = ySailPlanHeadLeft[iI]-fSailGradientLeft[iI]*xSailWindHeadLeft[iI] ;
      }
    if (xSailWindHeadRight[iI]==xSailWindFootRight[iI]) 
      {
      fSailGradientRight[iI] = 0 ;
      fSailInterceptRight[iI] = 0 ;
      }
    else
      {
      fSailGradientRight[iI] = (ySailPlanHeadRight[iI]-ySailPlanFootRight[iI])/(xSailWindHeadRight[iI]-xSailWindFootRight[iI]) ;
      fSailInterceptRight[iI] = ySailPlanHeadRight[iI]-fSailGradientRight[iI]*xSailWindHeadRight[iI] ;
      }
    }

  // Calculate the equations which define the upper and lower left edges of spanker seen from the direction of the wind
  iI = iSpanker ;
  fSailGradientLeft[iI] = (ySailPlanHeadLeft[iI]-ySailPlanHeadRight[iI])/(xSailWindHeadLeft[iI]-xSailWindHeadRight[iI]) ;
  fSailInterceptLeft[iI] = ySailPlanHeadLeft[iI]-fSailGradientLeft[iI]*xSailWindHeadLeft[iI] ;
  fSailGradient2[iI] = 0 ;
  fSailIntercept2[iI] = 0 ;

  // Calculate the equations which define the lower left edges of staysails seen from the diretion of the wind
  for (iI = 0; iI < iNrSails; iI++)
    {
    if ((iSailType[iI] == 1) || (iSailType[iI] == 3))
      {
      if (xSailWindFootLeft[iI]==xSailWindFootRight[iI]) 
        {
        fSailGradient2[iI] = 0 ;
        fSailIntercept2[iI] = 0 ;
        }
      else
        {
        fSailGradient2[iI] = (ySailPlanFootLeft[iI]-ySailPlanFootRight[iI])/(xSailWindFootLeft[iI]-xSailWindFootRight[iI]) ;
        fSailIntercept2[iI] = ySailPlanFootLeft[iI]-fSailGradient2[iI]*xSailWindFootLeft[iI] ;
        }
      }
    }

  // Now for each set sail...
  for (cSl=0 ; cSl < iNrSails; cSl++) 
    {
    fTotalPerpArea = 0.0 ;
    fEffPerpArea = 0.0 ;
    fSailExposedFract[cSl] = 0 ;
    if (iSailStatus[cSl] < 4)
      {
      // ...calculate the perpendicular area in bands of SAILSTEP feet
      for (fH = fSailEffectiveHead[cSl]-SAILSTEP/2;  fH >= fSailFoot[cSl]+SAILSTEP/2; fH = fH - SAILSTEP)
        {
        // first the total perpendicular width of the band
        if (fH > fSailCorner[cSl])
          {
          if ((fSailGradientLeft[cSl] != 0) && (fSailInterceptLeft[cSl] != 0))
            fLeftEdge = (fH-fSailInterceptLeft[cSl])/fSailGradientLeft[cSl] ;
          else
            fLeftEdge = xSailWindFootLeft[cSl] ;
          }
        else
          {
          if ((fSailGradient2[cSl] != 0) && (fSailIntercept2[cSl] != 0))
            fLeftEdge = (fH-fSailIntercept2[cSl])/fSailGradient2[cSl] ;
          else
            fLeftEdge = xSailWindFootLeft[cSl] ;
          }
        if ((fSailGradientRight[cSl] != 0) && (fSailInterceptRight[cSl] != 0))
          fRightEdge = (fH-fSailInterceptRight[cSl])/fSailGradientRight[cSl] ;
        else
          fRightEdge = xSailWindFootRight[cSl] ;
        if (fLeftEdge > fRightEdge)
          {
          fTemp = fLeftEdge ;
          fLeftEdge = fRightEdge ;
          fRightEdge = fTemp ;
          }
        fTotalPerpArea = fTotalPerpArea + fabs(fRightEdge - fLeftEdge) * SAILSTEP ;

        // ...and now the width of shadowing
  Again:
        bSplit = FALSE ;
        iN = 0 ;
        while (iSailCanShad[iSeq][cSl][iN] > -1)
          {
          cSh = iSailCanShad[iSeq][cSl][iN] ;
          if ((fH >= fSailFoot[cSh]) && (fH <= fSailEffectiveHead[cSh]))
            {
            // may shadow at this height 
            if (fH > fSailCorner[cSh])
              {
              if ((fSailGradientLeft[cSh] != 0) && (fSailInterceptLeft[cSh] != 0))
                fShadLeftEdge = (fH-fSailInterceptLeft[cSh])/fSailGradientLeft[cSh] ;
              else
                fShadLeftEdge = xSailWindFootLeft[cSh] ;
              }
            else
              {
              if ((fSailGradient2[cSh] != 0) && (fSailIntercept2[cSh] != 0))
                fShadLeftEdge = (fH-fSailIntercept2[cSh])/fSailGradient2[cSh] ;
              else
                fShadLeftEdge = xSailWindFootLeft[cSh] ;
              }
            if ((fSailGradientRight[cSh] != 0) && (fSailInterceptRight[cSh] != 0))
              fShadRightEdge = (fH-fSailInterceptRight[cSh])/fSailGradientRight[cSh] ;
            else
              fShadRightEdge = xSailWindFootRight[cSh] ;
            if (fShadLeftEdge > fShadRightEdge)
              {
              fTemp = fShadLeftEdge ;
              fShadLeftEdge = fShadRightEdge ;
              fShadRightEdge = fTemp ;
              }
            if ((fShadLeftEdge > fLeftEdge) && (fShadRightEdge < fRightEdge))
              // no shadow on both edges
              {
              fRightEdge = fShadLeftEdge ;
              bSplit = TRUE ;
              fNewLeftEdge = fShadRightEdge ;
              }
            if ((fShadLeftEdge <= fLeftEdge) && (fShadRightEdge > fLeftEdge))
              // shadow overlaps left edge
              fLeftEdge = fShadRightEdge ;
            if ((fShadLeftEdge < fRightEdge) && (fShadRightEdge >= fRightEdge))            
              //shadow overlaps right edge
              fRightEdge = fShadLeftEdge ;
            if (fLeftEdge > fRightEdge)
              // shadow overlaps both edges, band is totally shadowed
              {
              fRightEdge = fLeftEdge ;
              // no point in continuing
              if (! bSplit)
                break ;
              } 
            }
          iN = iN + 1 ;
          }

        fEffPerpArea = fEffPerpArea + fabs(fRightEdge - fLeftEdge) * SAILSTEP ;
        if (bSplit == TRUE)
          {
          fLeftEdge = fNewLeftEdge ;
          if ((fSailGradientRight[cSl] != 0) && (fSailInterceptRight[cSl] != 0))
            fRightEdge = (fH-fSailInterceptRight[cSl])/fSailGradientRight[cSl] ;
          else
            fRightEdge = xSailWindFootRight[cSl] ;
          goto Again ;
          }
        }
      }
    fSailPerpArea[cSl] = fEffPerpArea ;
    if (fTotalPerpArea < 0.000001)
      fSailExposedFract[cSl] = 0.0 ;
    else
      fSailExposedFract[cSl] = fEffPerpArea / fTotalPerpArea ;
    // correct for heeling
    fSailPerpArea[cSl] = fSailPerpArea[cSl] * cos(fHeelingAngle * fDegToRad) ;
    }


  for (iK=0; iK<iAcceleration[iSpeedFactor]; iK++)
    {
    // Calculate force acting on each sail and foreward, lateral and tortional thrust
    fForwardThrust = 0.0 ;
    fLeewardThrust = 0.0 ;
    fRotationalThrust = 0.0 ;

    fMoment = 0.0 ;
    fArea = 0.0  ; // to calculate COEForeAft

    for (iI = 0; iI< iNrSails; iI++)
      {
      iSailWarning[iI] = 0 ;
      if (iSailStatus[iI] > 3)
        {
        // furled or blown out
        fSailPerpFract[iI] = 0.0 ;
        iSailWarning[iI] = 4 ;
        }
      else
        {   
        if ((iSailType[iI] == 1) || (iSailType[iI] == 3) || (iI == iSpanker))
          // fore=and-aft sails
          fAngle = -fSailAngle[iI] - fAppWindAngle ;
         else 
           fAngle = (90.0 - fSailAngle[iI]) - fAppWindAngle ;
        // useful fraction
        fSailPerpFract[iI] = sin(fAngle * fDegToRad) ;
        // check if sail is ashiver and/or aback
        iSign = Sgn(fSailPerpFract[iI]) ;
        if ((iSailType[iI] == 1) || (iSailType[iI] == 3) || (iI == iSpanker))
          {
          if (fabs(fSailPerpFract[iI]) < 0.173) 
            {
            // sail ashiver 
            iSailWarning[iI] = 1 ; 
            // reduced efficiency at < 10  (arbitrary cubic function)
            fTemp = fSailPerpFract[iI] ;
            fSailPerpFract[iI] = iSign * fTemp * fTemp * fTemp * fTemp / 0.00518 ;
            }
          }
        else 
          { 
          if (fSailPerpFract[iI] < 0.5)
            {
            if (fSailPerpFract[iI] > 0.0)
              { 
              // sail ashiver
              iSailWarning[iI] = 1 ;
              // reduced efficiency at < 30  (arbitrary cubic function)
              fTemp = fSailPerpFract[iI] ;
              fSailPerpFract[iI] = iSign * fTemp * fTemp * fTemp * fTemp / 0.125 ;
              }
            else
              {  
              if (fSailPerpFract[iI] > -0.173)
                {
                // sail ashiver and aback
                iSailWarning[iI] = 2 ;
                // reduced efficiency at < 10  (arbitrary cubic function)
                fTemp = fSailPerpFract[iI] ;
                fSailPerpFract[iI] = iSign * fTemp * fTemp * fTemp * fTemp / 0.00518 ;
                }
              else
                {
                // sail aback
                iSailWarning[iI] = 3;
                }
              }  
            }
          }
        } 

      // useful force
      fSailPerpForce[iI] = fSailPerpArea[iI] * fAppWindSpeed * fAppWindSpeed * fSailPerpFract[iI] / 1000.0 ;

      fAngle = fSailAngle[iI] * fDegToRad ;
      if ((iSailType[iI] == 1) || (iSailType[iI] == 3) || (iI == iSpanker))
        // fore-and-aft sails (for which 0= fore-and-aft)
        fAngle = fPi / 2 - fAngle ;
      // useful force for headway; + = ahead
      fSailHeadwayForce[iI] = fSailPerpForce[iI] * cos(fAngle) ;
      if ((iSailType[iI] == 1) || (iSailType[iI] == 3) || (iI == iSpanker))
        fSailHeadwayForce[iI] = -fSailHeadwayForce[iI] ;
      // Useful force for leeway; +=to starboard 
      fSailLeewayForce[iI] = fSailPerpForce[iI] * sin(fAngle) ;
      fForwardThrust = fForwardThrust + fSailHeadwayForce[iI] ;
      fLeewardThrust = fLeewardThrust + fSailLeewayForce[iI] ;
      switch (iSailType[iI])
        {
        case 2:
          // foremast sail
          fDistance = fMastDistance[0] ;
          break ;
        case 4:
          // mainmast sail
          if (iI==iSpanker)
            fDistance = fMastDistance[1] - (fSailFootWidth[iSpanker] / 2.0) ;
          else
            fDistance = fMastDistance[1] ;
          break ;
        case 6:
          // mizzenmast sail
          if (iI==iSpanker)
            fDistance = fMastDistance[2] - (fSailFootWidth[iSpanker] / 2.0) ;
          else
            fDistance = fMastDistance[2] ;
          break ;
        case 1:
        case 3:
          // staysail
          fDistance = fSailSSCentroid[iI-(iSpanker+1)] ;
        }
    
      // Rotational force
      fRotatForce[iI] = fSailLeewayForce[iI] * (fDistance - fCORDistance) ;
      fRotationalThrust = fRotationalThrust + fRotatForce[iI] ;
      fMoment = fMoment + fSailLeewayForce[iI] * (fDistance - fCORDistance) ;
      fArea = fArea + fSailLeewayForce[iI] ;
      }

    // Do these calculations for the rudder
    fSailPerpArea[iNrSails] = fabs(fRudderArea * sin(iHelmAngle * fDegToRad)); // effective area
    fSailPerpFract[iNrSails] = fabs(sin(iHelmAngle * fDegToRad)); // useful fraction
    fSailPerpForce[iNrSails] = fSailPerpArea[iNrSails] * fSpeed * fSpeed * 
      fSailPerpFract[iNrSails] * fRudderFactor / 1000.0 ; // useful force
    fSailHeadwayForce[iNrSails] = -Sgn(fSpeed) * fabs(fSailPerpForce[iNrSails] *
      sin(iHelmAngle * fDegToRad)) ; // useful force for headway
    fSailLeewayForce[iNrSails] = Sgn(fSpeed) * Sgn(iHelmAngle) * fSailPerpForce[iNrSails] *
      cos(iHelmAngle * fDegToRad) ; // 'useful force for leeway
    fRotatForce[iNrSails] = -fSailLeewayForce[iNrSails] * ((fShipLength / 2.0)-fCORDistance) ;
    fForwardThrust = fForwardThrust + fSailHeadwayForce[iNrSails] ;
    fLeewardThrust = fLeewardThrust + fSailLeewayForce[iNrSails] ;
    fRotationalThrust = fRotationalThrust + fRotatForce[iNrSails] ;

    // Do these calculations for the hull and superstructure
    fSailPerpArea[iNrSails+1] = fabs(fHullSide * sin(fAppWindAngle * fDegToRad)); // effective area
    fSailPerpFract[iNrSails+1] = fabs(sin(fAppWindAngle * fDegToRad)); // useful fraction
    fSailPerpFract[iNrSails+1] = sin(-fAppWindAngle * fDegToRad) ; // <=======
    fSailPerpForce[iNrSails+1] = fSailPerpArea[iNrSails+1] * fAppWindSpeed * fAppWindSpeed * 
      fSailPerpFract[iNrSails+1] * fHullFactor / 1000.0 ; // useful force
    fAngle = fPi / 2 ;
    fSailHeadwayForce[iNrSails+1] = fSailPerpForce[iNrSails+1] * cos(fAngle) ;
    fSailHeadwayForce[iNrSails+1] = -fSailHeadwayForce[iNrSails+1] ;
      // Useful force for leeway; +=to starboard 
    fSailLeewayForce[iNrSails+1] = fSailPerpForce[iNrSails+1] * sin(fAngle) ;
    //fSailHeadwayForce[iNrSails+1] = fSailPerpForce[iNrSails+1] * cos(fAppWindAngle * fDegToRad) ; // useful force for headway
    //fSailLeewayForce[iNrSails+1] = -Sgn(fAppWindAngle) * fSailPerpForce[iNrSails+1] * sin(fAppWindAngle * fDegToRad) ; // 'useful force for leeway
    fRotatForce[iNrSails+1] = fSailLeewayForce[iNrSails+1] * (fHullCentroid-fCORDistance) ;
    fForwardThrust = fForwardThrust + fSailHeadwayForce[iNrSails+1] ;
    fLeewardThrust = fLeewardThrust + fSailLeewayForce[iNrSails+1] ;
    fRotationalThrust = fRotationalThrust + fRotatForce[iNrSails+1] ;
    fMoment = fMoment + fSailLeewayForce[iNrSails+1] * fHullCentroid;
    fArea = fArea + fSailLeewayForce[iNrSails+1];

    //Calculate position of Center of Effort
    if (fArea != 0)
      fCOEForeAft = fMoment / fArea ;
    else
      fCOEForeAft = -500.0 ;

    // Calculate heeling thrust
    fHeelingThrust = 0 ;
    fCOEHeight = 0 ;
    for (iI = 0; iI < iNrSails ; iI++)
      {
      fSailHeelForce[iI]= 0.0 ;
      if (iSailStatus[iI] < 4)
        {
        fSailHeelForce[iI] = fSailMeanHeight[iI] * cos(fHeelingAngle * fDegToRad) * fSailLeewayForce[iI] / 1000.0 ;
        fHeelingThrust = fHeelingThrust + fSailHeelForce[iI] ;
        fCOEHeight = fCOEHeight + fSailLeewayForce[iI] ;
        }
      }
    if (fCOEHeight != 0)
      fCOEHeight = 1000.0 * fHeelingThrust / fCOEHeight ;
    if (fabs(fCOEHeight - iLastCOEHeight) > 5)
      {
      iLastCOEHeight = (int) fCOEHeight ;
      bSternViewIsChanged = TRUE ;
      }

    // Calculate angle of heel

    fRightingMoment = -Sgn(fHeelingAngle) * fabs(sin(fHeelingAngle * 2.0 * fDegToRad) * 
      fShipMass * fHeelAngleFactor * fMaxHeelMoment) ;
    fResistance = -fHeelResistFactor * Sgn(fDeltaHeel) * fHeelSpeed * fHeelSpeed ;

    if (fabs(fHeelingThrust) < 0.01)
      {
      fResistance = 0 ;
      }

    fTotalThrust = fHeelingThrust + fRightingMoment + fResistance ;
    fDeltaSpeed = fTotalThrust * fIncrement / fVerticalInertia ;

    // damp excessive heeling acceleration (more than about 2/sec/sec)
    if (fabs(fDeltaSpeed) > 1.0)
      {
      fDeltaSpeed = 1.0 * Sgn(fDeltaSpeed) ;
      bAlertHeel = TRUE ;
      }
    else
      bAlertHeel = FALSE ;

    fHeelSpeed = fHeelSpeed + fDeltaSpeed ;
    fDeltaHeel = fHeelSpeed * fIncrement ;
    fHeelingAngle = fHeelingAngle + fDeltaHeel ;

    if ((Sgn((float)iLastHeelingAngle)!=Sgn(fHeelingAngle)) && (fHeelingThrust==0) ) 
      fHeelingAngle=0;
  
    if (fabs(fHeelingAngle) > 80.)
      fHeelingAngle = Sgn(fHeelingAngle)*80.0;

    if ((int) fHeelingAngle != iLastHeelingAngle)
      {
      iLastHeelingAngle = (int) fHeelingAngle ;
      bSternViewIsChanged = TRUE ;
      }

    // Calculate the contribution of heel to rotational forces

    fCOELeeward = -fCOEHeight * tan(fHeelingAngle * fDegToRad) ;
    // + = COE is to port
    if ((int) fCOELeeward != iLastCOELeeward)
      {
      iLastCOELeeward = (int) fCOELeeward ;
      bPlanViewIsChanged = TRUE ;
      }
    fEffectOfHeel = fCOELeeward * fForwardThrust ;
    fRotationalThrust = fRotationalThrust + fEffectOfHeel ;


    // Calculate the forward, leeward and rotational drags

    // Leeward

    fLeewardDrag = (fLeewaySpeed * fLeewaySpeed) * fCSASide * fWaterDragFactor / 2000.0 ; 
    //fLeewardDrag = (fLeewaySpeed * fLeewaySpeed) * (fCSASide * fWaterDragFactor * (1 + fabs(fSpeed)) / 2000.0) ;
    // arbitrary additional increase of drag with speed to reduce leeway
    fLeewardDrag = -fLeewardDrag * Sgn(fLeewaySpeed) ;
    if (fabs(fLeewardDrag) > (10.0 * fabs(fLeewardThrust)))
      fLeewardDrag = Sgn(fLeewardDrag) * 10.0 * fabs(fLeewardThrust) ;

    fDeltaSpeed = (fLeewardThrust + fLeewardDrag) * fIncrement / fShipMass ;

    // damp excessive acceleration to leeway (more than about 2 kt/sec)
    if (fabs(fDeltaSpeed) > 1.0)
      {
      fDeltaSpeed = 1.0 * Sgn(fDeltaSpeed) ;
      bAlertLeeway = TRUE ;
      } 
    else
      bAlertLeeway = FALSE ;

    fLeewaySpeed = fLeewaySpeed + fDeltaSpeed ;

    if (fabs(fLeewardThrust) < 0.5)
      {
      fLeewardDrag = 0 ;
      fLeewaySpeed = 0 ;
      }

    if (fabs(fSpeed) > 0.1)
      fLeewayAngle = Sgn(fSpeed) * atan(fLeewaySpeed / fSpeed) * 180.0 / fPi ;
    else
      fLeewayAngle = Sgn(fLeewaySpeed) * 90.0 ;

    // Forward

    if (fSpeed>0)
      fForwardDrag = (fSpeed * fSpeed) * fCSABow * fWaterDragFactor / 1000.0 ;
    else
      fForwardDrag = (fSpeed * fSpeed) * fCSAStern * fWaterDragFactor / 1000.0 ;
    fForwardDrag = -fForwardDrag * Sgn(fSpeed) ;

    fDeltaSpeed = (fForwardThrust + fForwardDrag) * fIncrement / fShipMass ;
    // damp excessive acceleration (more than about 2 kt/sec)
    if (fabs(fDeltaSpeed) > 1.0)
      {
      fDeltaSpeed = 1.0 * Sgn(fDeltaSpeed) ;
      bAlertSpeed = TRUE ;
      } 
    else
      bAlertSpeed = FALSE ;

    fSpeed = fSpeed + fDeltaSpeed ;

    //...rotational

    fLastRotationalThrust = fRotationalThrust ;

    fRotationalDrag = fShipLength * fShipLength * fShipLength * fShipLength * fShipDraught * 
      fRotationalSpeed * fRotationalSpeed * fRotatDragFactor / (32.0 * 1000.0) ;
    fRotationalDrag = -fRotationalDrag * Sgn(fRotationalSpeed) ;

    if (fabs(fRotationalDrag) > (10.0 * fabs(fRotationalThrust)))
      fRotationalDrag = Sgn(fRotationalDrag) * 10.0 * fabs(fRotationalThrust) ;
    fDeltaSpeed = (fRotationalThrust + fRotationalDrag) * fIncrement / fRotationalInertia ;
    fRotationalSpeed = fRotationalSpeed + fDeltaSpeed ;

    if (fRotationalThrust == 0.0)
      {
      fRotationalDrag = 0 ;
      fRotationalSpeed = 0 ;
      }

    fDeltaHdg = fRotationalSpeed * fIncrement * 180.0 / fPi ;
  
    // damp excessive rotational speed (more than about 2/sec)
    if (fabs(fDeltaHdg) > 0.2)
      {
      fRotationalSpeed = 0 ;
      fDeltaHdg = 0.2 * Sgn(fDeltaHdg) ;
      bAlertRotat = TRUE ;
      } 
    else
      bAlertRotat = FALSE ;

    //write forces to Sails-array to simplify display procedure
    fSailHeadwayForce[iNrSails+2] = fForwardThrust ;
    fSailHeadwayForce[iNrSails+3] = fForwardDrag ;
    fSailHeadwayForce[iNrSails+4] = fForwardThrust + fForwardDrag ;
    fSailLeewayForce[iNrSails+2] = fLeewardThrust ;
    fSailLeewayForce[iNrSails+3] = fLeewardDrag ;
    fSailLeewayForce[iNrSails+4] = fLeewardThrust + fLeewardDrag ;
    fRotatForce[iNrSails+5] = fEffectOfHeel ;
    fRotatForce[iNrSails+2] = fRotationalThrust ;
    fRotatForce[iNrSails+3] = fRotationalDrag ;
    fRotatForce[iNrSails+4] = fRotationalThrust + fRotationalDrag ;
    fSailHeelForce[iNrSails+2] = fHeelingThrust ;
    fSailHeelForce[iNrSails+3] = fRightingMoment ;
    fSailHeelForce[iNrSails+4] = fHeelingThrust + fRightingMoment + fResistance ;

    // Calculate new heading
    fDegreesPerMin = fDeltaHdg * 60.0 / fIncrement ;
    fHeading = fHeading + fDeltaHdg ;
    if (fHeading > 360.0)
      fHeading = fHeading - 360.0 ;
    if (fHeading <= 0.0)
      fHeading = fHeading + 360.0 ;

    // Check for blown out sails
    for (cSl = 0; cSl < iNrSails ;cSl++)
      {
      if ((iSailStatus[cSl] < 4) && (fSailPerpArea[cSl] != 0.0))
        {
        if (fAppWindSpeed > (fSailMaxSpeed[cSl] / fSetFraction[iSailStatus[cSl]])) 
          {
          if (!bPlayBack)
            {
            iSailStatus[cSl] = 5 ;
            sprintf(sMessage, "%s blown out", sSailName[cSl]) ; 
            ShowMessage () ; 
            if (bRecording)
              Recorder (9, cSl, 5) ;
            fSailMaxSpeed[cSl] = fSailTheorMaxSpeed[cSl] * (0.9 + 0.2 * (float) rand()/RAND_MAX) ;
            if (! bParalax)
              bSailPlanIsChanged = TRUE ;
              CanShadow () ;
            if (cSl > iSpanker)
              bPlanViewIsChanged = TRUE ; // remove staysail
            }
          }
        }
      } 

    // Calculate track and new position
    fSpeedNorth = fSpeed * cos(fHeading * fDegToRad) + fLeewaySpeed * cos((fHeading+90.) * fDegToRad) ;
    fSpeedEast = fSpeed * sin(fHeading * fDegToRad) + fLeewaySpeed * sin((fHeading+90.) * fDegToRad) ;
    fTrack = (atan(fSpeedEast/fSpeedNorth))/fDegToRad ;
    if (fSpeedEast < 0.0)
      {
      if (fTrack < 0.0)
        fTrack = 360 + fTrack ;
      else
        fTrack = 180 + fTrack ;
      }
    else
      if (fTrack < 0.0)
        fTrack = 180 + fTrack ;
    fLatitude = fLatitude + fSpeedNorth * fIncrement / 3600.0 ;
    fLongitude = fLongitude + fSpeedEast * fIncrement / 3600.0 ;

    // update recording or replay time  
    if (bRecording)
      {
      if (iLastRecTime != (int) fSimTime)
        {
        iLastRecTime = (int) fSimTime ;
        strcpy (sMessage, sMyProgName) ;
        strcat (sMessage, " - recording voyage ( ") ;
        strncat (sMessage, sTitleName2, 30) ;
        strcat (sMessage, ")") ;
        iHours = (iLastRecTime/3600) ;
        iMinutes = (iLastRecTime-(iHours*3600))/60 ;
        iSeconds = iLastRecTime - ((iHours*3600)+(iMinutes*60)) ;
        if ((iHours==0) && (iMinutes==0))
          sprintf(sBuf, "  %d sec", iSeconds) ;
        else
          if (iHours==0)
            sprintf(sBuf, "  %d min %d sec", iMinutes, iSeconds) ;
          else
            sprintf(sBuf, "  %d hr %d min %d sec", iHours, iMinutes, iSeconds) ;
        strcat (sMessage, sBuf) ;
        SetWindowText (hwnd, sMessage) ;
        }
      } 
    if (bPlayBack)
      {
      if (iLastRecTime != (int) fSimTime)
        {
        iLastRecTime = (int) fSimTime ;
        strcpy (sMessage, sMyProgName) ;
        strcat (sMessage, " - replaying voyage ( ") ;
        strncat (sMessage, sTitleName2, 30) ;
        strcat (sMessage, ")") ;
        iHours = (iLastRecTime/3600) ;
        iMinutes = (iLastRecTime-(iHours*3600))/60 ;
        iSeconds = iLastRecTime - ((iHours*3600)+(iMinutes*60)) ;
        if ((iHours==0) && (iMinutes==0))
          sprintf(sBuf, "  %d sec", iSeconds) ;
        else
          if (iHours==0)
            sprintf(sBuf, "  %d min %d sec", iMinutes, iSeconds) ;
          else
            sprintf(sBuf, "  %d hr %d min %d sec", iHours, iMinutes, iSeconds) ;
        strcat (sMessage, sBuf) ;
        SetWindowText (hwnd, sMessage) ;
        }
      // carry out any scheduled playback events
      while (fSimTime >= fEventTime [0])
        {
        switch (iEvent[0])
          {
          case 1: 
            // helm angle 
            iNewHelmAngle = iEventCode[0];
            ShowSelectedHelmAngle();
            if (bModernHelm)
              SetScrollPos (hwndSlider[4], SB_CTL, -iNewHelmAngle, TRUE) ;
            else
              SetScrollPos (hwndSlider[4], SB_CTL, iNewHelmAngle, TRUE) ;
            break ;
          case 2: 
            // yard angle 
            i1 = iEventCode[0] ;
            i2 = iEventCode[0] ;
            //check if yards are coupled
            if (bCoupleFore && (iEventCode[0] == 0 || iEventCode[0] == 1))
              {
              i1 = 0 ;
              i2 = 1 ;
              }
            if (bCoupleMizzen && (iEventCode[0] == 1 || iEventCode[0] == 2))
              {
              i1 = 1 ;
              i2 = 2 ;
              }
            if (bCoupleFore && bCoupleMizzen && (iEventCode[0] != 3))
              {
              i1 = 0 ;
              i2 = 2 ;
              }
            //adjust yard(s)
            for (iI=i1 ; iI<=i2 ; iI++)
              {
              iNewYardAngle[iI] = iEventCode2[0] ;
              if (iI==3)
                SetScrollPos (hwndSlider[iI], SB_CTL, -iNewYardAngle[iEventCode[0]], TRUE) ;
              else
                SetScrollPos (hwndSlider[iI], SB_CTL, iNewYardAngle[iEventCode[0]], TRUE) ;
              ShowSelectedYardAngle (iI) ;
              }
            break ;
          case 3: 
            // increase or decrease sail 
            iMadeSail = iEventCode[0] ;
            MakeSail () ;
            if (iMadeSail == 0)
              EnableWindow (hwndButton[3], FALSE) ;
            else
              EnableWindow (hwndButton[3], TRUE) ;
            if (iMadeSail == 15)
              EnableWindow (hwndButton[2], FALSE) ;
            else
              EnableWindow (hwndButton[2], TRUE) ;
            SetFocus (hwndCheck[1]) ;
            break ;
          case 4: 
            // set latitude 
            fDeltaLat = fLatitude-(iEventCode[0] + iEventCode2[0] / 100000.0) ;
            fLatitude = iEventCode[0] + iEventCode2[0] / 100000.0 ;
            break ;
          case 5: 
            // set longitude 
            fDeltaLon = fLongitude-(iEventCode[0] + iEventCode2[0] / 100000.0) ;
            fLongitude = iEventCode[0] + iEventCode2[0] / 100000.0 ;
              if ((fabs(fDeltaLat) > 0.0005) || (fabs(fDeltaLon) > 0.0005))
              {
              for (iI=iLastMapCounter+1; iI<=iMapCounter; iI++)
                {
                fLat[iI]=fLat[iI]-((float)(iI-iLastMapCounter)/(float)(iMapCounter-iLastMapCounter))*(fDeltaLat);
                fLon[iI]=fLon[iI]-((float)(iI-iLastMapCounter)/(float)(iMapCounter-iLastMapCounter))*(fDeltaLon);
                }
              bMapMustBePainted = TRUE ;
              }
            iLastMapCounter = iMapCounter ;
            break ;
          case 6: 
            // automatic yardsmen 
            if (iEventCode[0]==1)
              bYardsmen = FALSE ;
            else
              bYardsmen = TRUE ;
            Yardsmen () ;
            break ;
          case 7: 
            // automatic helmsman 
            if (iEventCode[0]==1)
              {
              fRequiredHeading = iEventCode2[0] ;
              bHelmsman = TRUE ;
              bEstablished = FALSE ;
              sprintf(sHelmInstruction, "steer %3.0f    ", fRequiredHeading); 
              SetUpDisplay () ;
              fAngle = fRequiredHeading - (fWindDirection + 180.0) ;
              if (fAngle < -180.0)
                fAngle = fAngle + 360.0 ;
              if (fAngle > 180.0)
                fAngle = fAngle - 360.0 ;
              if (fabs(fAngle) <= 105.0)
                {
                sprintf(sMessage, "Helmsman: steer %3.0f ", fRequiredHeading); 
                ShowMessage () ;
                }
              else
                {
                sprintf(sMessage, "Warning: heading of %3.0f  may be too close to wind", fRequiredHeading); 
                ShowMessage () ;
                }
              }
            else
              { 
              bHelmsman = TRUE ;
              Helmsman () ;
              }
            break ;
          case 8: 
            // jib/staysail angle 
            for (iI= iSpanker+1 ; iI < iNrSails ; iI++)
              {
              iSign = Sgn (fSailAngle[iI]) ;
              if (iSign == 0)
                iSign = Sgn (fAppWindAngle) ;
              if ((iEventCode[0] == 5) && (iSailType[iI]==1))
                fSailAngle[iI] = iSign * iEventCode2[0];
              if ((iEventCode[0] == 6) && (iSailType[iI]==3))
                fSailAngle[iI] = iSign * iEventCode2[0] ;
              if ((fabs(fSailAngle[iI]) > 44.0) && (iSailStatus[iI]==0))
                iSailStatus[iI] = 6; 
              if ((fabs(fSailAngle[iI]) < 44.0) && (iSailStatus[iI]==6))
                iSailStatus[iI] = 0; 
              }
            SetScrollPos (hwndSlider[iEventCode[0]], SB_CTL, iEventCode2[0], TRUE) ;
            if (iEventCode[0] == 5)
              ShowActualJibAngle();
            else
              ShowActualStaysailsAngle();
            bSailPlanIsChanged = TRUE ;
            break ;
          case 9: 
            // sail status
            ReplaySailStatus(iEventCode[0]);
            iSailStatus[iEventCode[0]] = iEventCode2[0] ;
            if (iEventCode[0] > iSpanker)
              bPlanViewIsChanged = TRUE ;
            SailPlan () ;
            CanShadow () ;
            break ;
          case 10: 
            // coupled mizzenyards 
            bCoupleMizzen = iEventCode[0] ;
            SendMessage (hwndCheck[1], BM_SETCHECK, bCoupleMizzen, 0);
            break ;
          case 11: 
            // coupled foreyards 
            bCoupleFore = iEventCode[0] ;
            SendMessage (hwndCheck[0], BM_SETCHECK, bCoupleFore, 0);
            break ;
          case 12: 
            // wind direction 
            if (iEventCode[0]==1)
              fNewWindDirection = iEventCode2[0] ;
            else
              {
              iWindDirIncr = iEventCode2[0] ;
              sprintf(sMessage, "Wind direction changing to %3.0f  at %3d /min", fNewWindDirection, iWindDirIncr); 
              ShowMessage () ;
              fDeltaWindDir = iWindDirIncr/120. ;
              bWindDirChange = TRUE ;
              }
            break ;
          case 13: 
            // wind speed 
            if (iEventCode[0]==1)
              fNewWindSpeed = iEventCode2[0] ;
            else
              {
              iWindSpeedIncr = iEventCode2[0] ;
              sprintf(sMessage, "Wind speed changing to %3.0f kts at %3d kts/min", fNewWindSpeed, iWindSpeedIncr); 
              ShowMessage () ;
              fDeltaWindSpeed = iWindSpeedIncr/120. ;
              bWindSpeedChange = TRUE ;
              }
            break ;
          case 99:
            EndPlayBack () ;
            break ;
          default:
            // replay item
            sprintf(sMessage, "Unrecognised event: %4.3f %2d", fEventTime[0], iEvent[0]) ; 
            ShowMessage () ;
          }
        if (bPlayBack)
          {
          for (iI = 0 ; iI < iLastReplayItem ; iI++)
            {
            fEventTime[iI] = fEventTime[iI+1] ; 
            iEvent[iI] = iEvent[iI+1] ;
            iEventCode[iI] = iEventCode[iI+1] ; 
            iEventCode2[iI] = iEventCode2[iI+1] ; 
            }
          iLastReplayItem-- ;
          if ((iLastReplayItem < 4) && (iEvent[iLastReplayItem]!= 99))
            Replayer () ;
          }
        else
          break ;
        }
      }
    // Update cycle count
    iCycles = iCycles +1 ;
    }
  }


/* --------------------------------------
   ReplaySailStatus: shows new sail status when replaying
   Version date = 03.07.19
----------------------------------------*/

void ReplaySailStatus(iThisSail)
  {
  int iThisStatus ;
  int iNrReefs ;

  iThisStatus = iSailStatus[iThisSail] ;
  iNrReefs = iSailMaxReefs[iThisSail] ;

  // sail was blown uit
  if (iThisStatus == 5)
    {
    sprintf(sMessage, "new %s bent", sSailName[iThisSail]) ; 
    ShowMessage () ; 
    return ;
    }

  // sail is now blown uit
  if (iEventCode2[0] == 5)
    {
    sprintf(sMessage, "%s blown out", sSailName[iThisSail]) ; 
    ShowMessage () ; 
    return ;
    }

  if (iNrReefs == 0) 
    {
    // sail is not reefable
    if (iThisStatus == 4)
      {
      sprintf(sMessage, "%s set", sSailName[iThisSail]) ; 
      ShowMessage () ; 
      }
    else
      {
      sprintf(sMessage, "%s furled", sSailName[iThisSail]) ; 
      ShowMessage () ; 
      }
    return ;
    }

  // sail is reefable
  if (iEventCode2[0] == 4)
    {
    // all reefs are already taken in; // furl
    sprintf(sMessage, "%s furled", sSailName[iThisSail]) ; 
    ShowMessage () ; 
    return ;
    } 
  if (iEventCode2[0] < iSailStatus[iThisSail])
    {
    if (iThisStatus == 4)
      {
      // sail was furled
      sprintf(sMessage, "%s set close reefed", sSailName[iThisSail]) ; 
      ShowMessage () ; 
      }
    else
      {
      sprintf(sMessage, "%s reef shaken out", sSailName[iThisSail]) ; 
      ShowMessage () ; 
      } 
    return ;
    }  
  sprintf(sMessage, "%s reefed", sSailName[iThisSail]) ; 
  ShowMessage () ; 

  }


